#include <stdio.h>
#include "game.h"
#include "control.h"
#include "civclient.h"
#include "government.h"
#define spiral_iterate(dx, dy, dz, r)\
int dx, dy, dz;\
for(dz = 0; dz <= r; dz++)\
  for(dy = -dz; dy <= dz; dy++)\
    for(dx = -dz; dx <= dz; dx++)\
      if((dy * dy) + (dx * dx) >= (dz * dz))
extern struct connection aconnection;
static int mapped = 0;

static int reachable(int x, int y, struct map_position *desti)
{
if(map_get_terrain(desti->x, desti->y) != T_UNKNOWN)
  {
  if(map_get_continent(desti->x, desti->y))	/* not ocean */
    {
    spiral_iterate(dx, dy, dz, 1)
      if(map_same_continent(desti->x, desti->y, map_adjust_x(x + dx), y - dy))
        return(1);
    }
  if(map_get_continent(x, y))	/* not ocean */
    {
    spiral_iterate(dx, dy, dz, 1)
      if(map_same_continent(map_adjust_x(desti->x + dx), desti->y - dy, x, y))
        return(1);
    }
  }
return(x == desti->x && y == desti->y);
}

static struct map_position *can_board(struct unit *punit, int radius)
{
if(is_ground_unit(punit) && unit_flag(punit->type, F_SETTLERS))
  {
  static struct map_position desti;
  spiral_iterate(dx, dy, dz, radius)
    if((dz || !radius) && is_transporter_with_free_space(game.player_ptr,
		desti.x = map_adjust_x(punit->x + dx), desti.y = punit->y - dy)
        && reachable(punit->x, punit->y, &desti))
      return(&desti);
  }
return(NULL);
}

static struct map_position *expand(int x, int y, int radius)
{
spiral_iterate(dx, dy, dz, radius)
  {
  static struct map_position desti;
  desti.x = map_adjust_x(x + dx);
  desti.y = y - dy;
    {
    int overlap = 0;
    enum tile_terrain_type pterr = map_get_terrain(desti.x, desti.y);
    spiral_iterate(ex, ey, ez, 2)
      if(map_get_city(map_adjust_x(desti.x + ex), desti.y - ey))
        overlap++;
    if(!overlap && pterr != T_UNKNOWN && pterr != T_OCEAN
	&& reachable(x, y, &desti))
      return(&desti);
    }
  }
return(NULL);
}

static struct map_position *enemy_unit(int x, int y, int radius)
{
spiral_iterate(dx, dy, dz, radius)
  {
  static struct map_position desti;
  desti.x = map_adjust_x(x + dx);
  desti.y = y - dy;
  if(reachable(x, y, &desti) && map_get_terrain(desti.x, desti.y) != T_OCEAN)
    unit_list_iterate(map_get_tile(desti.x, desti.y)->units, iunit)
      if(iunit->owner != game.player_ptr->player_no)
        return(&desti);
    unit_list_iterate_end;
  }
return(NULL);
}

static struct map_position *enemy_city(struct unit *punit, int radius)
{
spiral_iterate(dx, dy, dz, radius)
  {
  spiral_iterate(ex, ey, ez, radius > 1 ? 1 : 0)
    {
    static struct map_position desti;
    struct city *pcity = map_get_city(
	map_adjust_x((desti.x = map_adjust_x(punit->x + dx)) + ex),
	(desti.y = punit->y - dy) - ey);
//if(pcity)fprintf(stderr, "%d, %d : %d, %d : %d, %d %s\n", punit->x, punit->y, desti.x, desti.y, pcity->x, pcity->y, pcity->name);
    if(pcity && pcity->owner != game.player_ptr->player_no
	&& reachable(punit->x, punit->y, &desti))
        return(&desti);
    }
  }
return(NULL);
}

static int has_boat(struct city *pcity)
{
unit_list_iterate(game.player_ptr->units, iunit)
  if(unit_has_role(iunit->type, L_FERRYBOAT) && iunit->homecity == pcity->id)
    return(1);
unit_list_iterate_end;
return(0);
}

static int passenger(struct unit *punit)
{
int count = 0;
if(unit_has_role(punit->type, L_FERRYBOAT))
  unit_list_iterate(map_get_tile(punit->x, punit->y)->units, iunit)
    if(is_ground_unit(iunit) && iunit->activity == ACTIVITY_SENTRY)
      count++;
  unit_list_iterate_end;
return(count);
}


static int defenders(int x, int y)
{
int count = 0;
unit_list_iterate(map_get_tile(x, y)->units, iunit)
  if(iunit->owner == game.player_ptr->player_no
	&& (unit_has_role(iunit->type, L_DEFEND_OK)
	|| unit_has_role(iunit->type, L_DEFEND_GOOD)))
    count++;
unit_list_iterate_end;
return(count);
}

static struct map_position *guarded(int x, int y, int radius)
{
static struct map_position desti;
spiral_iterate(dx, dy, dz, radius)
  if(dz && defenders(desti.x = map_adjust_x(x + dx), desti.y = y - dy)
	&& !map_get_city(desti.x, desti.y) && reachable(x, y, &desti))
    return(&desti);
return(NULL);
}

static Unit_Type_id best_attacker(struct city *pcity)
{
int strength = 0;
Unit_Type_id select = 0, type;
for(type = 0; type < game.num_unit_types; type++)
  if(can_build_unit(pcity, type) && unit_types[type].attack_strength > strength)
    {
    strength = unit_types[type].attack_strength;
    select = type;
    }
return(select);
}

static Unit_Type_id best_defender(struct city *pcity)
{
int str = 0;
Unit_Type_id select = 0, type;
for(type = 0; type < game.num_unit_types; type++)
  if(can_build_unit(pcity, type) && unit_types[type].defense_strength > str)
    {
    str = unit_types[type].defense_strength;
    select = type;
    }
return(select);
}

static Unit_Type_id best_build(struct city *pcity)
{
return(
    enemy_unit(pcity->x, pcity->y, 2)
      ? (defenders(pcity->x, pcity->y)
        ? best_attacker(pcity)
        : best_defender(pcity))
      : !is_transporter_with_free_space(game.player_ptr, pcity->x, pcity->y)
                && best_role_unit(pcity, L_FERRYBOAT) != U_LAST
                && !has_boat(pcity)
        ? best_role_unit(pcity, L_FERRYBOAT)
        : expand(pcity->x, pcity->y, 4)
          ? best_role_unit(pcity, F_SETTLERS)
          : defenders(pcity->x, pcity->y) > 2 || mapped || pcity->shield_stock
                > unit_types[best_role_unit(pcity, F_SETTLERS)].build_cost
            ? (guarded(pcity->x, pcity->y, 6)
              ? best_attacker(pcity)
              : best_defender(pcity))
            : best_role_unit(pcity, F_SETTLERS)
      );
}

static int have_tech(char *target)
{
static Tech_Type_id previous = A_LAST;
struct packet_player_request packet;
Tech_Type_id techid = find_tech_by_name(target);
enum packet_type method = 
  get_invention(game.player_ptr, techid) == TECH_REACHABLE
    ? PACKET_PLAYER_RESEARCH : PACKET_PLAYER_TECH_GOAL;
if(get_invention(game.player_ptr, techid) == TECH_KNOWN
	|| game.player_ptr->research.researching == techid)
  {
  previous = techid;
  return(1);
  }
packet.tech = techid;
if(game.player_ptr->ai.tech_goal != techid
	&& !(game.player_ptr->research.researching == previous
		&& method == PACKET_PLAYER_RESEARCH))
  send_packet_player_request(&aconnection, &packet, method);
return(0);
}

int aiplay(void)
{
int call_again = 0;
static int newcity = 0;
static int lastyear = 0;
struct unit *punit = get_unit_in_focus();
struct city *pcity = punit ? map_get_city(punit->x, punit->y) : NULL;
struct map_position *pdest;
if(lastyear != game.year)
  {
  newcity = 0;
  if(game.player_ptr->government == find_government_by_name("Anarchy")->index
	&& !game.player_ptr->revolution)
    {
    struct packet_player_request packet;
    packet.tax = 30;
    packet.luxury = 0;
    packet.science = 70;
    packet.government = find_government_by_name("Monarchy")->index;
    send_packet_player_request(&aconnection, &packet, PACKET_PLAYER_GOVERNMENT);
    send_packet_player_request(&aconnection, &packet, PACKET_PLAYER_RATES);
    }
  if(game.player_ptr->government == game.default_government)
    {
    struct packet_player_request packet;
    packet.tax = 40;
    packet.luxury = 0;
    packet.science = 60;
    if(game.government_count)
      if(can_change_to_government(game.player_ptr,
		find_government_by_name("Monarchy")->index))
        send_packet_player_request(
		&aconnection, &packet, PACKET_PLAYER_REVOLUTION);
      else
        send_packet_player_request(&aconnection, &packet, PACKET_PLAYER_RATES);
    }
  if(have_tech("Alphabet"))
    if(have_tech("Map Making"))
      if(have_tech("Monarchy"))
        if(have_tech("Mathematics"))
          if(have_tech("Navigation"))
            have_tech("Feudalism");
  city_list_iterate(game.player_ptr->cities, icity)
    {
    struct packet_city_request packet;
    packet.name[0] = '\0';
    packet.city_id = icity->id;
    packet.is_build_id_unit_id = TRUE;
    packet.build_id = best_build(icity);
    if(packet.build_id != icity->currently_building)
      send_packet_city_request(&aconnection, &packet, PACKET_CITY_CHANGE);
    if((enemy_unit(icity->x, icity->y, 2) && !defenders(icity->x, icity->y))
		|| icity->shield_stock * 10
			> unit_types[icity->currently_building].build_cost * 9)
      send_packet_city_request(&aconnection, &packet, PACKET_CITY_BUY);
    if(city_unhappy(icity))
      {
      int x, y;
      city_map_iterate_outwards(x, y)
        if(is_worker_here(icity, x, y))
          {
          packet.worker_x = x;
          packet.worker_y = y;
          }
      city_map_iterate_outwards_end;
      send_packet_city_request(
		&aconnection, &packet, PACKET_CITY_MAKE_SPECIALIST);
      }
    }
  city_list_iterate_end;
  unit_list_iterate(game.player_ptr->units, iunit)
    {
    request_unit_wakeup(iunit);
    if(iunit->activity == ACTIVITY_GOTO && !enemy_unit(iunit->x, iunit->y, 1))
      request_new_unit_activity(iunit, ACTIVITY_IDLE);
    if(unit_has_role(iunit->type, L_FERRYBOAT) && !passenger(iunit)
	&& iunit->activity == ACTIVITY_EXPLORE)
      request_new_unit_activity(iunit, ACTIVITY_IDLE);
    if(unit_has_role(iunit->type, L_FERRYBOAT)
	&& iunit->activity == ACTIVITY_EXPLORE
	&& iunit->moves_left == unit_types[iunit->type].move_rate)
      mapped = 1;
    }
  unit_list_iterate_end;
  }
if(!punit)
  key_end_turn();
else if(unit_has_role(punit->type, L_EXPLORER) || passenger(punit))
  key_unit_explore();
else if(unit_has_role(punit->type, L_FERRYBOAT)
	&& !passenger(punit) && !map_get_city(punit->x, punit->y))
  {
  pcity = find_city_by_id(punit->homecity);
  request_unit_goto();
  do_goto(pcity->x, pcity->y);
  }
else if(is_military_unit(punit) && (pdest = enemy_city(punit, 1))
	&& !enemy_unit(pdest->x, pdest->y, 0))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
  }
else if(unit_flag(punit->type, F_SETTLERS) && expand(punit->x, punit->y, 0)
	&& (punit->y * map.xsize) + punit->x + 1 != newcity)
  {
  static int count = 0;
  struct packet_unit_request req;
  req.unit_id = punit->id;
  sprintf(req.name, "City%d", count++);
  send_packet_unit_request(&aconnection, &req, PACKET_UNIT_BUILD_CITY);
  newcity = (punit->y * map.xsize) + punit->x + 1;
  }
else if(can_board(punit, 0) && (pcity
	|| count_terrain_near_tile(punit->x, punit->y, T_OCEAN) == 8))
  key_unit_sentry();
else if(unit_flag(punit->type, F_SETTLERS)
	&& (pdest = expand(punit->x, punit->y, 4))
	&& !enemy_unit(pdest->x, pdest->y, 1)
	&& !enemy_unit(punit->x, punit->y, 1))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
//  fprintf(stderr, "Expand: %s(%d, %d) moves to %s(%d, %d)\n", unit_name(punit->type), punit->x, punit->y, get_tile_type(map_get_tile(pdest->x, pdest->y)->terrain)->terrain_name, pdest->x, pdest->y);
  }
else if((pdest = can_board(punit, 4))
	&& map_get_terrain(punit->x, punit->y) != T_OCEAN
	&& !enemy_unit(punit->x, punit->y, 1))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
//  fprintf(stderr, "Board : %s(%d, %d) moves to %s(%d, %d)\n", unit_name(punit->type), punit->x, punit->y, get_tile_type(map_get_tile(pdest->x, pdest->y)->terrain)->terrain_name, pdest->x, pdest->y);
  }
else if(pcity && defenders(pcity->x, pcity->y) <= 2
	&& (unit_has_role(punit->type, L_DEFEND_GOOD)
	|| unit_has_role(punit->type, L_DEFEND_OK)))
  key_unit_fortify();
else if((!pcity || defenders(pcity->x, pcity->y) > 2)
	&& (unit_has_role(punit->type, L_DEFEND_GOOD)
	|| unit_has_role(punit->type, L_DEFEND_OK))
	&& (pdest = enemy_city(punit, 6)))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
  fprintf(stderr, "Guard: %s(%d, %d) moves to %s(%d, %d)\n", unit_name(punit->type), punit->x, punit->y, get_tile_type(map_get_tile(pdest->x, pdest->y)->terrain)->terrain_name, pdest->x, pdest->y);
  }
else if(unit_types[punit->type].attack_strength
	> unit_types[punit->type].defense_strength
        && !unit_has_role(punit->type, L_FERRYBOAT)
	&& (pdest = enemy_unit(punit->x, punit->y, 2)))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
  fprintf(stderr, "Attack : %s(%d, %d) moves to %s(%d, %d)\n", unit_name(punit->type), punit->x, punit->y, get_tile_type(map_get_tile(pdest->x, pdest->y)->terrain)->terrain_name, pdest->x, pdest->y);
  }
else if(unit_types[punit->type].attack_strength
	> unit_types[punit->type].defense_strength
        && !unit_has_role(punit->type, L_FERRYBOAT)
	&& (pdest = guarded(punit->x, punit->y, 6)))
  {
  request_unit_goto();
  do_goto(pdest->x, pdest->y);
  fprintf(stderr, "Follow: %s(%d, %d) moves to %s(%d, %d)\n", unit_name(punit->type), punit->x, punit->y, get_tile_type(map_get_tile(pdest->x, pdest->y)->terrain)->terrain_name, pdest->x, pdest->y);
  }
else if(enemy_city(punit, 1)
	&& (unit_has_role(punit->type, L_DEFEND_GOOD)
	|| unit_has_role(punit->type, L_DEFEND_OK)))
  key_unit_fortify();
else if(unit_flag(punit->type, F_SETTLERS) && mapped)
  key_unit_auto();
else
  {
  key_unit_done();
  call_again = 1;
  }
lastyear = game.year;
return(call_again);
}
